﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// Helper class that calculates polynomial hash, which allows for O(1) comparison of arbitrary sub-lists in a causality chain.
    /// </summary>
    static class PolyHash
    {
        private const long factor = 1542691;

        /// <summary>
        /// Adds one element to the hash
        /// </summary>
        /// <param name="aggregate">Current hash value</param>
        /// <param name="next">New element value</param>
        /// <returns>New hash value</returns>
        public static long Append(long aggregate, long next)
        {
            return unchecked(aggregate * factor + next);
        }

        /// <summary>
        /// Calculates the hash for a list interval, given the hashes of super-list and sub-list sharing an end, and difference between their cardinality.
        /// E.g. given a super-list of elements [N downto 0] and a sub-list [M downto 0], where N >= M, and cardinality difference (N - M), returns a hash for interval [N downto M+1].
        /// </summary>
        /// <param name="higher">Hash of a super-list</param>
        /// <param name="lower">Hash of a sub-list</param>
        /// <param name="steps">Cardinality difference</param>
        /// <returns>Hash of the interval</returns>
        public static long Subtract(long higher, long lower, int steps)
        {
            unchecked
            {
                for (int i = 0; i < steps; i++)
                    lower *= factor;
                return higher - lower;
            }
        }
    }
}
