﻿using FlowPreservation;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace FlowPreservation
{
    /// <summary>
    /// A graph representation for causality chains
    /// </summary>
    class StackGraph
    {
        private const string ns = "{http://schemas.microsoft.com/vs/2009/dgml}";

        private static readonly Regex singleFrameRegex = new Regex(@"(" + StackFrameSlim.AsyncBoundary + "[\r\n]+)?.*[\r\n]+");
        private static readonly Regex indentRegex = new Regex(@"^\s*");
        private static readonly Regex multiLineRegex1 = new Regex(@"\s*(" + StackFrameSlim.FirstArrow + @"|:(?=line))\s*");
        private static readonly Regex multiLineRegex2 = new Regex(@"\s*" + StackFrameSlim.SecondArrow + @"\s*");

        private readonly Dictionary<string, StackGraphNode> nodes = new Dictionary<string, StackGraphNode>();

        /// <summary>
        /// Runs through all frames in the segment and, if equal frame exists already, reuses the node, otherwise creates a new one
        /// </summary>
        /// <param name="concatenated">Concatenated causality chain for one thread, each line containing a separate frame</param>
        public void AddSegment(string concatenated)
        {
            var singleFrames = singleFrameRegex.Matches(concatenated);
            StackGraphNode previous = null;
            foreach (var singleFrame in singleFrames.Cast<Match>().Select(m => m.Value))
            {
                string replacement = Environment.NewLine + "    " + indentRegex.Match(singleFrame).Value;
                var frame = multiLineRegex1.Replace(singleFrame, replacement);
                if (frame == singleFrame)
                    frame = multiLineRegex2.Replace(singleFrame, replacement);

                StackGraphNode current;
                if (!nodes.TryGetValue(frame, out current))
                {
                    current = new StackGraphNode(frame);
                    nodes.Add(frame, current);
                }

                if (previous != null)
                    current.Links.Add(previous);

                previous = current;
            }
        }

        /// <summary>
        /// Collapses node chains to single nodes
        /// </summary>
        public void Collapse()
        {
            var parents = (
                from n in nodes.Values
                join l in (
                    from n in nodes.Values
                    from l in n.Links
                    select new { From = n, To = l }
                ) on n equals l.To into parentNodes
                select new { Node = n, Parent = parentNodes.Count() == 1 ? parentNodes.Single().From : null }
            ).ToDictionary(t => t.Node, t => t.Parent);

            foreach (var node in nodes.Values.ToArray())
            {
                var parent = parents[node];
                while (parent != null && !nodes.ContainsKey(parent.Frame))
                    parent = parents[parent];
                if (parent != null && parent.Links.Count() == 1)
                {
                    nodes.Remove(parent.Frame);
                    parent.Frame = node.Frame + parent.Frame;
                    parent.Links = node.Links;
                    nodes.Add(parent.Frame, parent);
                    nodes.Remove(node.Frame);
                }
            }
        }

        /// <summary>
        /// Graph vertices
        /// </summary>
        public IEnumerable<StackGraphNode> Nodes
        {
            get { return nodes.Values; }
        }

        /// <summary>
        /// Creates DGML xml content
        /// </summary>
        /// <returns>XDocument containing DGML</returns>
        public XDocument GetDgml()
        {
            var doc = new XDocument(
                new XElement(ns + "DirectedGraph", 
                    new XAttribute("GraphDirection", "BottomToTop"), 
                    new XAttribute("Layout", "Sugiyama"),
                    new XElement(ns + "Nodes", 
                        Nodes.Select(n => 
                            new XElement(ns + "Node", 
                                new XAttribute("Id", n.Id),
                                new XAttribute("Label", n.Frame))).ToArray()),
                    new XElement(ns + "Links", 
                        Nodes.SelectMany(n => n.Links.Select(l => 
                            new XElement(ns + "Link",
                                new XAttribute("Source", n.Id),
                                new XAttribute("Target", l.Id)))).ToArray())));
            return doc;
        }
    }
}
