﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// Lighter analog to StackFrame
    /// </summary>
    [DebuggerDisplay("{LongHash} : {ToString()}")]
    struct StackFrameSlim : IEquatable<StackFrameSlim>
    {
        internal const char FirstArrow = '\u21e6';
        internal const char SecondArrow = '\u2190';

        public static readonly StackFrameSlim AsyncBoundary = new StackFrameSlim(new IntPtr(-1L), -1, -1, null, -1, -1, false);
        private static readonly string asyncBoundaryString = "\u231B " 
            + string.Join(" ", Enumerable.Range(0, 10).Select(i => '\xB7'));

        public StackFrameSlim(IntPtr rgMethodHandle, int rgiOffset, int rgiILOffset,
            string rgFilename, int rgiLineNumber, int rgiColumnNumber, bool rgiLastFrameFromForeignExceptionStackTrace)
        {
            this.Method = default(MethodBaseSlim);

            this.MethodHandle = rgMethodHandle;
            this.Offset = rgiOffset;
            this.ILOffset = rgiILOffset;
            this.FileName = rgFilename;
            this.LineNumber = rgiLineNumber;
            this.ColumnNumber = rgiColumnNumber;
            this.IsLastFrameFromForeignExceptionStackTrace = rgiLastFrameFromForeignExceptionStackTrace;
        }

        public MethodBaseSlim Method;

        public readonly IntPtr MethodHandle;
        public readonly int Offset;
        public readonly int ILOffset;
        public readonly string FileName;
        public readonly int LineNumber;
        public readonly int ColumnNumber;
        public readonly bool IsLastFrameFromForeignExceptionStackTrace;

        public override string ToString()
        {
            var builder = new StringBuilder();
            if (Equals(AsyncBoundary))
                builder.Append(asyncBoundaryString);
            else
            {
                builder.Append((Equals(Method, default(MethodBaseSlim)) ? null : Method.StringValue) ?? "<UnknownFrame>");
                if (!string.IsNullOrEmpty(FileName) || LineNumber != 0 || ColumnNumber != 0)
                    builder.AppendFormat(" {3} {0}:line {1}:column {2}", FileName, LineNumber, ColumnNumber, FirstArrow);
                //builder.AppendFormat(" {2} 0x{0:X} bytes (0x{1:X} IL)", Offset, ILOffset, SecondArrow);
            }
            return builder.ToString();
        }

        /// <summary>
        /// Hashes are long to reduce collisions when relying on hash comparisons between stack trace segments
        /// </summary>
        public long LongHash
        {
            get
            {
                return unchecked(MethodHandle.ToInt64()
                    ^ ((uint)ILOffset * 76319L)
                    ^ ((uint)Offset * 1743920413L)
                    ^ (FileName == null ? 0 : (uint)FileName.GetHashCode() * 9432173L)
                    ^ (LineNumber * 18472183L)
                    ^ (ColumnNumber * 29428193823821L)
                    ^ (IsLastFrameFromForeignExceptionStackTrace ? 0x400000 : 0));
            }
        }

        public override int GetHashCode()
        {
            long hash = LongHash;
            return (int)(hash ^ (hash >> 32));
        }

        public override bool Equals(object obj)
        {
            if (obj == null || obj.GetType() != typeof(StackFrameSlim))
                return false;

            return Equals((StackFrameSlim)obj);
        }

        public bool Equals(StackFrameSlim other)
        {
            return MethodHandle == other.MethodHandle
                && Offset == other.Offset
                && ILOffset == other.ILOffset
                && FileName == other.FileName
                && LineNumber == other.LineNumber
                && ColumnNumber == other.ColumnNumber
                && IsLastFrameFromForeignExceptionStackTrace == other.IsLastFrameFromForeignExceptionStackTrace;
        }

        /// <summary>
        /// Checks if the frame should be visible against justMyCode filter
        /// </summary>
        /// <param name="justMyCode">Hide framework boilerplate</param>
        /// <returns>Whether frame needs to be hidden</returns>
        public bool IsHidden(bool justMyCode)
        {
            return !Equals(AsyncBoundary) && (Method.IsEventInfrastructure || justMyCode && Method.IsExternalCode);
        }
    }
}
