﻿using FlowPreservation;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using System.Runtime.Remoting.Messaging;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace FlowPreservationPerfTests
{
    /// <summary>
    /// Simple set of performance tests
    /// </summary>
    public class PerfTests
    {
        async public Task TestAsyncLocalPerf()
        {
            const int iterations = 10000;

            await MeasurePerfAsync(async () => await Task.Yield(), "await Task.Yield()", iterations);

            await MeasurePerfAsync(async () => await Task.Yield(), "await Task.Yield()", iterations);

            CallContext.LogicalSetData("Temp", "Hello!");
            await MeasurePerfAsync(async () => await Task.Yield(), "await Task.Yield() with CallContext data", iterations);
            Console.WriteLine("Data: " + CallContext.LogicalGetData("Temp"));
        }

        public void TestStackTracePerf()
        {
            const int iterations = 1000;

            MeasurePerf(() => new StackTraceSlim(false), "new StackTraceSlim(false)", iterations);
            MeasurePerf(() => new StackTraceSlim(true), "new StackTraceSlim(true)", iterations);
            MeasurePerf(() => new StackTrace(false), "new StackTrace(false)", iterations);
            MeasurePerf(() => new StackTrace(true), "new StackTrace(true)", iterations);
        }

        async public Task TestStackTracePerfYield()
        {
            const int iterations = 1000;

            await MeasurePerfAsync(async () => { await Task.Yield(); }, "await Task.Yield()", iterations);
            await MeasurePerfAsync(async () => { await Task.Yield(); new StackTraceSlim(false); }, "await Task.Yield(); new StackTraceSlim(false)", iterations);
            await MeasurePerfAsync(async () => { await Task.Yield(); new StackTraceSlim(true); }, "await Task.Yield(); new StackTraceSlim(true)", iterations);
            await MeasurePerfAsync(async () => { await Task.Yield(); new StackTrace(false); }, "await Task.Yield(); new StackTrace(false)", iterations);
            await MeasurePerfAsync(async () => { await Task.Yield(); new StackTrace(true); }, "await Task.Yield(); new StackTrace(true)", iterations);
        }

        async public Task TestStackTracePerfAwaitSync()
        {
            const int iterations = 1000;

            await MeasurePerfAsync(async () => { await Task.FromResult(0); }, "await Task.FromResult(0)", iterations);
            await MeasurePerfAsync(async () => { await Task.FromResult(0); new StackTraceSlim(false); }, "await Task.FromResult(0); new StackTraceSlim(false)", iterations);
            await MeasurePerfAsync(async () => { await Task.FromResult(0); new StackTraceSlim(true); }, "await Task.FromResult(0); new StackTraceSlim(true)", iterations);
            await MeasurePerfAsync(async () => { await Task.FromResult(0); new StackTrace(false); }, "await Task.FromResult(0); new StackTrace(false)", iterations);
            await MeasurePerfAsync(async () => { await Task.FromResult(0); new StackTrace(true); }, "await Task.FromResult(0); new StackTrace(true)", iterations);
        }

        async public Task TestStoragePerf()
        {
            const int iterations = 1000;
            var methods = new Func<Task>[] { InnerNoAssert, MiddleNoAssert, OuterNoAssert };

            FlowReservoir.Enroll();

            foreach (var task in methods)
                await MeasurePerfAsync(task, task.Method.Name + " with logging", iterations);
            foreach (var task in methods)
                await MeasurePerfAsync(task, task.Method.Name + " with logging", iterations);
            foreach (var task in methods)
                await MeasurePerfAsync(task, task.Method.Name + " with logging", iterations);
        }

        private void MeasurePerf(Action action, string name, int iterations)
        {
            Task completedTask = Task.FromResult((object)null);
            MeasurePerfAsync(() => { action(); return completedTask; }, name, iterations).Wait();
        }

        async private Task MeasurePerfAsync(Func<Task> action, string name, int iterations)
        {
            await action();
            var stopwatch = Stopwatch.StartNew();
            for (int i = 0; i < iterations; i++)
                await action();
            stopwatch.Stop();
            string toWrite = name + ": " + stopwatch.Elapsed;
            Console.WriteLine(toWrite);
            Trace.WriteLine(toWrite);
        }

        async private static Task OuterNoAssert()
        {
            await MiddleNoAssert();
        }

        async private static Task MiddleNoAssert()
        {
            await InnerNoAssert();
        }

        async private static Task InnerNoAssert()
        {
            await Task.Yield();
        }
    }
}
