﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// Wraps a void-returning ICriticalNotifyCompletion-based awaiter to execute an action 
    /// after asynchronous part is completed, but before it schedules the continuation
    /// </summary>
    /// <typeparam name="TAwaiter">Awaiter type</typeparam>
    public struct CriticalAppendableAwaiter<TAwaiter> : ICriticalNotifyCompletion where TAwaiter : ICriticalNotifyCompletion
    {
        private readonly TAwaiter awaiter;
        private readonly Action preCompleted;
        private readonly Func<TAwaiter, bool> isCompleted;
        private readonly Action<TAwaiter> getResult;

        public CriticalAppendableAwaiter(TAwaiter awaiter, Action preCompleted, 
            Func<TAwaiter, bool> isCompleted, Action<TAwaiter> getResult)
        {
            this.awaiter = awaiter;
            this.preCompleted = preCompleted;
            this.isCompleted = isCompleted;
            this.getResult = getResult;
        }

        public bool IsCompleted 
        { 
            get { return isCompleted(awaiter); } 
        }

        public void UnsafeOnCompleted(Action continuation)
        {
            var thisCopy = this;
            awaiter.OnCompleted(thisCopy.preCompleted.Append(continuation));
        }

        public void OnCompleted(Action continuation) 
        {
            var thisCopy = this;
            awaiter.OnCompleted(thisCopy.preCompleted.Append(continuation));
        }

        public void GetResult() 
        { 
            getResult(awaiter); 
        }
    }

    /// <summary>
    /// Wraps a result-returning ICriticalNotifyCompletion-based awaiter to execute an action 
    /// after asynchronous part is completed, but before it schedules the continuation
    /// </summary>
    /// <typeparam name="TAwaiter">Awaiter type</typeparam>
    /// <typeparam name="TResult">Result type</typeparam>
    public struct CriticalAppendableAwaiter<TAwaiter, TResult> : ICriticalNotifyCompletion where TAwaiter : ICriticalNotifyCompletion
    {
        private readonly TAwaiter awaiter;
        private readonly Action preCompleted;
        private readonly Func<TAwaiter, bool> isCompleted;
        private readonly Func<TAwaiter, TResult> getResult;

        public CriticalAppendableAwaiter(TAwaiter awaiter, Action preCompleted,
            Func<TAwaiter, bool> isCompleted, Func<TAwaiter, TResult> getResult)
        {
            this.awaiter = awaiter;
            this.preCompleted = preCompleted;
            this.isCompleted = isCompleted;
            this.getResult = getResult;
        }

        public bool IsCompleted
        {
            get { return isCompleted(awaiter); }
        }

        public void UnsafeOnCompleted(Action continuation)
        {
            var thisCopy = this;
            awaiter.OnCompleted(thisCopy.preCompleted.Append(continuation));
        }

        public void OnCompleted(Action continuation)
        {
            var thisCopy = this;
            awaiter.OnCompleted(thisCopy.preCompleted.Append(continuation));
        }

        public TResult GetResult()
        {
            return getResult(awaiter);
        }
    }
}
