﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// Abstract base class for async-local storage, allowing for event listeners to access thread-local storages
    /// </summary>
    public abstract class AsyncLocal
    {
        [ThreadStatic]
        internal static Dictionary<Guid, object> PersistentThreadStorage;

        [ThreadStatic]
        internal static Dictionary<Guid, object> DiscardableThreadStorage;

        protected AsyncLocal()
        {
            AsyncLocalListener.EnsureInitialized();
        }
    }

    /// <summary>
    /// Generic concrete class for async-local storage 
    /// </summary>
    /// <typeparam name="T">Type of value stored in async-local storage</typeparam>
    public class AsyncLocal<T> : AsyncLocal
    {
        private readonly Guid guid = Guid.NewGuid();
        private readonly Func<T> valueFactory;
        private readonly bool discardOnAwaitBoundary;

        public AsyncLocal() : this(() => default(T)) { }

        /// <summary>
        /// Constructor, allowing to instantiate two different flavors of async-local storage,
        /// where the value is either discarded on first await boundary or preserved along the execution flow
        /// </summary>
        /// <param name="valueFactory"></param>
        /// <param name="discardOnAwaitBoundary"></param>
        public AsyncLocal(Func<T> valueFactory, bool discardOnAwaitBoundary = false)
        {
            this.valueFactory = valueFactory;
            this.discardOnAwaitBoundary = discardOnAwaitBoundary;
        }

        public T Value
        {
            get 
            {
                object value;
                if (ThreadStorage.TryGetValue(guid, out value))
                    return (T)value;
                else
                {
                    var newValue = valueFactory();
                    ThreadStorage.Add(guid, newValue);
                    return newValue;
                }
            }
            set 
            {
                ThreadStorage[guid] = value; 
            }
        }

        private Dictionary<Guid, object> ThreadStorage
        {
            get 
            {
                if (discardOnAwaitBoundary)
                    return DiscardableThreadStorage ?? (DiscardableThreadStorage = new Dictionary<Guid, object>());
                else
                    return PersistentThreadStorage ?? (PersistentThreadStorage = new Dictionary<Guid, object>());
            }
        }
    }
}
