﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.Tracing;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// TplEtwProvider and FrameworkEventSource listener supporting the implementation of AsyncLocal{T}
    /// </summary>
    class AsyncLocalListener : TplFrameworkListenerBase
    {
        private static AsyncLocalListener listener;
        public static void EnsureInitialized()
        {
            EnsureInitialized(() => new AsyncLocalListener(), ref listener);
        }

        private readonly ConcurrentDictionary<int, Dictionary<Guid, object>> temporaryStorage =
            new ConcurrentDictionary<int, Dictionary<Guid, object>>();

        protected override void TaskWaitBegin(int taskId, EventConstants.Tpl.TaskWaitBehavior behavior)
        {
            if (!temporaryStorage.TryAdd(taskId, AsyncLocal.PersistentThreadStorage))
                throw new AsyncFlowDiagnosticsException("Data already exists in temporary storage for task " + taskId);
        }

        /// <summary>
        /// Only transferring values to the new thread for PersistentThreadStorage, 
        /// i.e. the one used by AsyncLocal{T} inititialized with discardOnAwaitBoundary = false
        /// </summary>
        protected override void TaskWaitEnd(int taskId)
        {
            Dictionary<Guid, object> threadStorage;
            if (temporaryStorage.TryRemove(taskId, out threadStorage))
                AsyncLocal.PersistentThreadStorage = threadStorage;
            else
                throw new AsyncFlowDiagnosticsException("Data not present in temporary storage for task " + taskId);
        }

        /// <summary>
        /// New unrelated work item starts executing on current thread - need to clean up thread-local storage
        /// </summary>
        protected override void ResetLocal()
        {
            AsyncLocal.PersistentThreadStorage = null;
            AsyncLocal.DiscardableThreadStorage = null;
        }
    }
}
