﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;

namespace FlowPreservation
{
    /// <summary>
    /// Main class that is used to preserve causality information in Windows Store applications
    /// </summary>
    public static class FrameStorage
    {
        private static readonly AsyncLocal<LinkedList<string>> Frames =
            new AsyncLocal<LinkedList<string>>(() => new LinkedList<string>(Enumerable.Repeat(string.Empty, 1)), true);

        private static readonly AsyncLocal<StrongBox<LinkedListNode<string>>> PreviousSegmentHead =
            new AsyncLocal<StrongBox<LinkedListNode<string>>>(() => new StrongBox<LinkedListNode<string>>(), true);

        #region WithCausality

        /// <summary>
        /// Adds currently executing method to frames list and registers an action to remove it 
        /// as soon as asynchronous work completes
        /// </summary>
        /// <param name="task">Void-returning task</param>
        public static Task WithCausality(this Task task, 
            [CallerMemberName] string member = null, [CallerFilePath] string file = null, [CallerLineNumber] int line = 0)
        {
            var removeAction = AddAndCreateRemoveAction(member, file, line);
            return task.ContinueWith(t => removeAction());
        }

        /// <summary>
        /// Adds currently executing method to frames list and registers an action to remove it 
        /// as soon as asynchronous work completes
        /// </summary>
        /// <param name="task">Result-returning task</param>
        public static Task<T> WithCausality<T>(this Task<T> task, 
            [CallerMemberName] string member = null, [CallerFilePath] string file = null, [CallerLineNumber] int line = 0)
        {
            var removeAction = AddAndCreateRemoveAction(member, file, line);
            return task.ContinueWith(t => { removeAction(); return t.Result; });
        }

        /// <summary>
        /// Adds currently executing method to frames list and registers an action to remove it 
        /// as soon as asynchronous work completes
        /// </summary>
        /// <param name="awaitable">An awaitable returned by Task.Yield()</param>
        public static CriticalAppendableAwaitable<YieldAwaitable.YieldAwaiter> WithCausality(this YieldAwaitable awaitable, 
            [CallerMemberName] string member = null, [CallerFilePath] string file = null, [CallerLineNumber] int line = 0)
        {
            var removeAction = AddAndCreateRemoveAction(member, file, line);
            return new CriticalAppendableAwaitable<YieldAwaitable.YieldAwaiter>(() => awaitable.GetAwaiter(), removeAction, 
                awaiter => awaiter.IsCompleted, awaiter => awaiter.GetResult());
        }

        /// <summary>
        /// Adds currently executing method to frames list and registers an action to remove it 
        /// as soon as asynchronous work completes
        /// </summary>
        /// <param name="awaitable">An awaitable returned by Task.ConfigureAwait() on a void-returning task</param>
        public static CriticalAppendableAwaitable<ConfiguredTaskAwaitable.ConfiguredTaskAwaiter> WithCausality(this ConfiguredTaskAwaitable awaitable, 
            [CallerMemberName] string member = null, [CallerFilePath] string file = null, [CallerLineNumber] int line = 0)
        {
            var removeAction = AddAndCreateRemoveAction(member, file, line);
            return new CriticalAppendableAwaitable<ConfiguredTaskAwaitable.ConfiguredTaskAwaiter>(() => awaitable.GetAwaiter(), removeAction,
                awaiter => awaiter.IsCompleted, awaiter => awaiter.GetResult());
        }

        /// <summary>
        /// Adds currently executing method to frames list and registers an action to remove it 
        /// as soon as asynchronous work completes
        /// </summary>
        /// <param name="awaitable">An awaitable returned by Task.ConfigureAwait() on a result-returning task</param>
        public static CriticalAppendableAwaitable<ConfiguredTaskAwaitable<T>.ConfiguredTaskAwaiter, T> WithCausality<T>(this ConfiguredTaskAwaitable<T> awaitable, 
            [CallerMemberName] string member = null, [CallerFilePath] string file = null, [CallerLineNumber] int line = 0)
        {
            var removeAction = AddAndCreateRemoveAction(member, file, line);
            return new CriticalAppendableAwaitable<ConfiguredTaskAwaitable<T>.ConfiguredTaskAwaiter, T>(() => awaitable.GetAwaiter(), removeAction,
                awaiter => awaiter.IsCompleted, awaiter => awaiter.GetResult());
        }

        #endregion

        /// <summary>
        /// Returns stored causality chain for current asynchronous operation (thread) except for currently executing frame
        /// </summary>
        public static string CausalityChain
        {
            get { return string.Join(Environment.NewLine, Frames.Value).Trim(); }
        }

        private static string FormatMemberInfo(string member, string file, int line)
        {
            return member + " in " + file + " at line " + line;
        }

        /// <summary>
        /// Adds frame to the causality linked list and returns an action to remove it
        /// </summary>
        /// <returns>An action that removes the frame just added from the causality list</returns>
        private static Action AddAndCreateRemoveAction(string member, string file, int line)
        {
            var framesValueCopy = Frames.Value;
            var previousSegmentHeadValueCopy = PreviousSegmentHead.Value;
            if (previousSegmentHeadValueCopy.Value == null)
                previousSegmentHeadValueCopy.Value = framesValueCopy.First;
            framesValueCopy.AddBefore(previousSegmentHeadValueCopy.Value, FormatMemberInfo(member, file, line));
            var addedNode = previousSegmentHeadValueCopy.Value.Previous;
            return () =>
            {
                PreviousSegmentHead.Value.Value = null;
                previousSegmentHeadValueCopy.Value = null;
                framesValueCopy.Remove(addedNode);
                Frames.Value = framesValueCopy;
            };
        }
    }
}
