﻿using EnvDTE;
using EnvDTE80;
using FlowPreservation;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace FlowViewer
{
    /// <summary>
    /// Extracts stored causality information for classic applications.
    /// </summary>
    class ClassicStackHelper : StackHelperBase
    {
        public ClassicStackHelper(DTE2 dte) : base(dte) { }

        /// <summary>
        /// Extracts causality information for the current thread by evaluating field expressions in the debugger, 
        /// reconstructing instances of classes representing stored causality using these values,
        /// and then calling their formatting methods directly.
        /// </summary>
        /// <returns>Stored causality chain except for the most recent stack segment, which is still on the stack</returns>
        public override string GetCausalityChain()
        {
            var reservoir = dte.Debugger.GetExpression(typeof(FlowReservoir).FullName);
            var stackStorage = reservoir.DataMember(typeof(FlowReservoir), typeof(StackStorage));
            var preservedStacks = stackStorage.WithStaticMembers()
                .DataMember(typeof(StackStorage), typeof(StackTraceNode));
            if (!IsNull(preservedStacks))
            {
                var stackTraceNode = CreateStackTraceNode(preservedStacks);
                return stackTraceNode.ToStringEx();
            }
            else
                return null;
        }

        private StackTraceNode CreateStackTraceNode(Expression expression)
        {
            var next = expression.DataMember((StackTraceNode n) => n.Next);
            return new StackTraceNode(
                CreateStackTraceSegment(expression.DataMember((StackTraceNode n) => n.Value)),
                IsNull(next) ? null : CreateStackTraceNode(next));
        }

        private StackTraceSegment CreateStackTraceSegment(Expression expression)
        {
            return new StackTraceSegment(
                CreateArray(expression.DataMember((StackTraceSegment s) => s.Frames), CreateStackFrameSlim),
                CreateArray(expression.DataMember((StackTraceSegment s) => s.Hashes), e => long.Parse(e.Value)),
                CreateArray(expression.DataMember((StackTraceSegment s) => s.Loops), CreateStackTraceLoop),
                int.Parse(expression.DataMember((StackTraceSegment s) => s.TotalCount).Value),
                int.Parse(expression.DataMember((StackTraceSegment s) => s.TotalCountWithLoops).Value));
        }

        private T[] CreateArray<T>(Expression expression, Func<Expression, T> createElement)
        {
            var array = new T[expression.DataMembers.Count];
            var indicesAndElements = Enumerable.Range(0, array.Length)
                .Zip(expression.DataMembers.Cast<Expression>(), (i, e) => new { i, e });
            foreach (var indexAndElement in indicesAndElements)
                array[indexAndElement.i] = createElement(indexAndElement.e);
            return array;
        }

        private StackFrameSlim CreateStackFrameSlim(Expression expression)
        {
            var stackFrameSlim = new StackFrameSlim(
                new IntPtr(int.Parse(expression.DataMember((StackFrameSlim f) => f.MethodHandle).Value)),
                int.Parse(expression.DataMember((StackFrameSlim f) => f.Offset).Value),
                int.Parse(expression.DataMember((StackFrameSlim f) => f.ILOffset).Value),
                StringParse(expression.DataMember((StackFrameSlim f) => f.FileName)),
                int.Parse(expression.DataMember((StackFrameSlim f) => f.LineNumber).Value),
                int.Parse(expression.DataMember((StackFrameSlim f) => f.ColumnNumber).Value),
                bool.Parse(expression.DataMember((StackFrameSlim f) => f.IsLastFrameFromForeignExceptionStackTrace).Value));
            stackFrameSlim.Method = CreateMethodBaseSlim(expression.DataMember((StackFrameSlim f) => f.Method));
            return stackFrameSlim;
        }

        private MethodBaseSlim CreateMethodBaseSlim(Expression expression)
        {
            if (IsNull(expression))
                return default(MethodBaseSlim);

            return new MethodBaseSlim(
                StringParse(expression.DataMember((MethodBaseSlim m) => m.StringValue)),
                bool.Parse(expression.DataMember((MethodBaseSlim m) => m.IsEventInfrastructure).Value),
                bool.Parse(expression.DataMember((MethodBaseSlim m) => m.IsExternalCode).Value));
        }

        private StackTraceLoop CreateStackTraceLoop(Expression expression)
        {
            return new StackTraceLoop(
                int.Parse(expression.DataMember((StackTraceLoop l) => l.HighIndex).Value),
                int.Parse(expression.DataMember((StackTraceLoop l) => l.LowIndex).Value),
                int.Parse(expression.DataMember((StackTraceLoop l) => l.Count).Value));
        }
    }
}
