﻿using EnvDTE;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FlowViewer
{
    /// <summary>
    /// Extensions class for crawling debugger expressions
    /// </summary>
    static class DteExtensions
    {
        private const int retries = 5;

        /// <summary>
        /// Given a debugger expression, returns a derived debugger expression
        /// </summary>
        /// <param name="expression">Original debugger expression</param>
        /// <param name="name">Name of member</param>
        /// <returns>New expression = original.name</returns>
        public static Expression DataMember(this Expression expression, string name)
        {
            if (expression == null)
                return null;

            Expression result;
            for (int i = 0; i < retries; i++)
            {
                result = expression.DataMembers.Cast<Expression>().SingleOrDefault(e => e.Name == name);
                if (result != null && (result.IsValidValue || result.Value.Contains("thread is stopped at a point where")))
                    return result;
            }
            return null;
        }

        /// <summary>
        /// Given a debugger expression, returns a derived debugger expression, 
        /// extracting member name from linq expression passed in
        /// </summary>
        /// <typeparam name="T">Declaring type</typeparam>
        /// <typeparam name="M">Member type</typeparam>
        /// <param name="expression">Original debugger expression</param>
        /// <param name="getter">Getter linq expression, from which to extract member name</param>
        /// <returns>New expression = original.name_of_the_member_accessed_by_the_getter</returns>
        public static Expression DataMember<T, M>(this Expression expression, System.Linq.Expressions.Expression<Func<T, M>> getter)
        {
            return expression.DataMember(ReflectionHelper.MemberName(getter));
        }

        /// <summary>
        /// Given a debugger expression, returns a derived debugger expression,
        /// finding member and its name by field type.
        /// </summary>
        /// <param name="expression">Original debugger expression</param>
        /// <param name="declaringType">Type of original debugger expression</param>
        /// <param name="fieldType">Type of the field defined on the declaring type</param>
        /// <returns>New expression = original.name_of_the_single_member_of_fieldType</returns>
        public static Expression DataMember(this Expression expression, Type declaringType, Type fieldType)
        {
            return expression.DataMember(ReflectionHelper.FieldName(declaringType, fieldType));
        }

        /// <summary>
        /// Given a debugger expression, returns a derived debugger expression,
        /// finding member by predicate.
        /// </summary>
        /// <param name="expression">Original debugger expression</param>
        /// <param name="predicate">Delegate that checks for a condition on debugger expression</param>
        /// <returns>Single derived expression that matches predicate</returns>
        public static Expression DataMember(this Expression expression, Func<Expression, bool> predicate)
        {
            return expression == null ? null : expression.DataMembers.Cast<Expression>().Single(predicate);
        }

        /// <summary>
        /// Given a debugger expression, returns an expression for its "Static members" node if it exists
        /// or original expression otherwise
        /// </summary>
        /// <param name="expression">Original debugger expression</param>
        /// <returns>Expression for "Static members" or the original one</returns>
        public static Expression WithStaticMembers(this Expression expression)
        {
            return expression.DataMember("Static members") ?? expression;
        }

        /// <summary>
        /// Given a debugger expression, returns an expression for its "Non-Public members" node 
        /// and optionally "Raw View" if they exist or original expression otherwise
        /// </summary>
        /// <param name="expression">Original debugger expression</param>
        /// <returns>Expression for "Non-Public members" and possibly "Raw View" or the original one</returns>
        public static Expression WithNonPublicMembers(this Expression expression)
        {
            expression = expression.DataMember("Raw View") ?? expression;
            return expression.DataMember("Non-Public members") ?? expression;
        }
    }
}
