﻿using EnvDTE;
using EnvDTE80;
using EnvDTE90a;
using FlowPreservation;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;

namespace FlowViewer
{
    /// <summary>
    /// Base class for plugins
    /// </summary>
    abstract class PluginBase
    {
        protected static readonly string eol = Environment.NewLine;
        protected readonly DTE2 dte;
        protected readonly AddIn addIn;
        protected readonly ClassicStackHelper classicStackHelper;
        protected readonly ModernStackHelper modernStackHelper;

        private readonly bool runOnlyWhenDebugging;
        private int inProgress = 0;

        public PluginBase(DTE2 dte, AddIn addIn, bool runOnlyWhenDebugging)
        {
            this.dte = dte;
            this.addIn = addIn;
            this.runOnlyWhenDebugging = runOnlyWhenDebugging;
            this.classicStackHelper = new ClassicStackHelper(dte);
            this.modernStackHelper = new ModernStackHelper(dte);
        }

        /// <summary>
        /// Main entry point
        /// </summary>
        async public void RunAsync()
        {
            await Task.Run(new Action(RunSync));
        }

        /// <summary>
        /// Runs the plugin, checking for existing running instance and debugging mode
        /// </summary>
        public void RunSync()
        {
            if (runOnlyWhenDebugging && (dte.Debugger.DebuggedProcesses == null || dte.Debugger.DebuggedProcesses.Count == 0
                || dte.Debugger.CurrentMode == dbgDebugMode.dbgRunMode))
            {
                dte.StatusBar.Text = "This action is only available in break mode during debugging";
                dte.StatusBar.Highlight(true);
                return;
            }

            if (Interlocked.CompareExchange(ref inProgress, 1, 0) != 0)
            {
                dte.StatusBar.Text = "Action is already in progress";
                dte.StatusBar.Highlight(true);
                return;
            }

            try
            {
                Run();
            }
            finally
            {
                Interlocked.Exchange(ref inProgress, 0);
            }
        }

        protected abstract void Run();

        /// <summary>
        /// Gets stored causality chain for both classic and Windows Store applications, 
        /// including the topmost stack segment, which is extracted directly from Call Stack window.
        /// </summary>
        /// <param name="thread">Thread to look up causality information for</param>
        /// <returns>Causality chain formatted as string</returns>
        protected string GetFullCausalityChain(EnvDTE.Thread thread)
        {
            var originalThread = dte.Debugger.CurrentThread;
            var originalFrame = dte.Debugger.CurrentStackFrame;
            dte.Debugger.CurrentThread = thread;
            try
            {
                var topSegment = StackHelperBase.GetTopStackSegmentAndSelectUserFrame(dte, true);

                var reservoir = dte.Debugger.GetExpression(typeof(FlowReservoir).FullName);
                var storedStack = !reservoir.Value.Contains("does not exist in the")
                    ? classicStackHelper.GetCausalityChain()
                    : modernStackHelper.GetCausalityChain();

                return topSegment + storedStack;
            }
            catch (COMException ex)
            {
                if ((uint)ex.ErrorCode == (uint)0x89711006) // No valid context in which to evaluate the expression. 
                    return null;
                else
                    throw;
            }
            catch (NullReferenceException)
            {
                return null;
            }
            finally
            {
                dte.Debugger.CurrentThread = originalThread;
                dte.Debugger.CurrentStackFrame = originalFrame;
            }
        }
    }
}
