﻿using EnvDTE;
using EnvDTE80;
using EnvDTE90a;
using FlowPreservation;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace FlowViewer
{
    /// <summary>
    /// Base class for classes that extract stored causality information 
    /// </summary>
    abstract class StackHelperBase
    {
        private static readonly Regex unresolvedFrameRegex = new Regex(@"^[0-9a-f]{8}(?:[0-9a-f]{8})?$");

        protected readonly DTE2 dte;

        public StackHelperBase(DTE2 dte)
        {
            this.dte = dte;
        }

        protected bool IsNull(Expression expression)
        {
            return expression == null || expression.Value == "null";
        }

        protected string StringParse(Expression expression)
        {
            if (IsNull(expression))
                return null;

            return expression.Value.Trim('"').Replace(@"\\", @"\");
        }

        /// <summary>
        /// Gets current stack segment for current thread by querying Call Stack window data
        /// and selects topmost user frame in the debugger, so that debugger expression
        /// evaluation has the right context
        /// </summary>
        /// <param name="dte">DTE</param>
        /// <param name="justMyCode">Whether non-user code frames should be skipped</param>
        /// <returns>Most recent stack segment, which is currently on stack</returns>
        public static string GetTopStackSegmentAndSelectUserFrame(DTE2 dte, bool justMyCode)
        {
            var topFrames = new List<string>();
            bool topUserFrameSelected = false;
            bool managedToNativeTransitionEncountered = false;
            foreach (var frame in dte.Debugger.CurrentThread.StackFrames.Cast<StackFrame2>())
            {
                string frameText = frame.FunctionName;
                if (frameText == "[Managed to Native Transition]")
                {
                    if (!managedToNativeTransitionEncountered)
                    {
                        topUserFrameSelected = false;
                        managedToNativeTransitionEncountered = true;
                    }
                    continue;
                }

                if (ExternalCodeHelper.IsEventInfrastructure(frameText)
                || justMyCode && (ExternalCodeHelper.IsExternalCode(frameText)
                    || frameText == "[Native to Managed Transition]"
                    || frameText.StartsWith("[Frames below may be incorrect and/or missing")
                    || unresolvedFrameRegex.IsMatch(frameText)))
                    continue;

                if (frameText == "[Resuming Async Method]")
                    frameText = StackFrameSlim.AsyncBoundary.ToString();
                else
                {
                    if (!frameText.StartsWith("[") || !frameText.EndsWith("]"))
                    {
                        frameText += "(" + string.Join(", ", frame.Arguments.Cast<Expression>().Select(a => a.Type + " " + a.Name)) + ")";
                        var fileName = frame.FileName;
                        if (!string.IsNullOrEmpty(fileName))
                            frameText += string.Format(" {0} {1}:line {2}", StackFrameSlim.FirstArrow, fileName, frame.LineNumber);
                    }
                }
                topFrames.Add(frameText);
                if (!topUserFrameSelected)
                {
                    topUserFrameSelected = true;
                    dte.Debugger.CurrentStackFrame = frame;
                }
            }
            while (topFrames.LastOrDefault() == StackFrameSlim.AsyncBoundary.ToString())
                topFrames.RemoveAt(topFrames.Count - 1);

            var builder = new StringBuilder();
            foreach (var frame in topFrames)
                builder.AppendLine(frame);
            return builder.ToString();
        }

        public abstract string GetCausalityChain();
    }
}
